// edge.h				Last updated: Feb 24, 98

#ifndef Edge_H
#define Edge_H

#include "common.h"

class Node;	// defined in graph.h

///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : UsedBand
//
//	Description :
//		Each edge has a list of used bandwidth by sessions. Every
//		edge in the graph information has a valid used bandwidth
//		information
//
///////////////////////////////////////////////////////////////////////////////
class UsedBand {
private:
	int	session_id;	// session ID
	int	used_bandwidth;	// one value per session
	UsedBand*	next;	// to form the list of used bandwidth items,
				// one item for every session
public:
	UsedBand(int s_id, int band) {	// init the attributes
		session_id = s_id;
		used_bandwidth = band;
		next = NIL;
	}

	UsedBand(const UsedBand* u);	// another way to init the attributes
					// see edge.cc
	int	GetSessID(void) const { return session_id; } // return session ID
	void	SetSessID(int s) { session_id = s; } 	     // set session ID to parm s

	int	GetBand(void) const { return used_bandwidth; } // return used bandwidth
	void	SetBand(int band) { used_bandwidth = band; }   // set used bandwidth to parm band
	int	IncBand(int band) {	// increase used bandwidth by parm band
		used_bandwidth += band;
		return used_bandwidth;
		}
	int	DecBand(int band) { 	// decrease used bandwidth by parm band
		used_bandwidth -= band;
		return used_bandwidth;
		}
	UsedBand* GetNext(void) const { return next; } // return next item in the list
	void	SetNext(UsedBand* n) { next = n; }     // set pointer next to n
};


///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : Edge
//
//	Description : 
//		An item of adjacency list for each node in the graph
//		Edge from source_node to target_node_id
//	Note:
//		An edge is connected by two nodes, the starting node
//		(start node) and the ending node (end node).
//
///////////////////////////////////////////////////////////////////////////////
class Edge {
private:
	Node* 	source_node;	// Start node reference
	int	target_node_id;	// End node ID
	int	edge_weight; 	// weight of the edge
	int	edge_delay;	// delay of the edge
	int	total_band;	// Total bandwidth capacity of the edge
	int	current_band;	// Available bandwidth for use on the edge
	UsedBand* used_band_head; // Head of list of used bandwidth items
	STATUS	edge_status;	// Used by SESSION class
				// LEAF_NODE (participant node), INTREE (intermediate node)
	float	edgeNormFactor; // edge normalization factor, used for sharedCost

public:
	Edge();
	Edge(Node* src, int t_node, int weight, int delay, int t_band);
	Edge(const Edge* edge);

	~Edge();

	float	GetEdgeNormFactor() const {return edgeNormFactor;}
	void 	SetEdgeNormFactor(float normFact) {edgeNormFactor = normFact;}

	Node*	GetSrcNode(void) const { return source_node; }  // return start node
	void	SetSrcNode(Node* nd) { source_node = nd; }	// set start node to parm nd

	int	GetTargetNodeID() const { return target_node_id; } // return end node ID
	void	SetTargetNodeID(int id) { target_node_id = id; }   // set end node ID to parm id

	int	GetWeight() const { return edge_weight; } // return edge ccost
	void	SetWeight(int w) { edge_weight = w; };	  // set ccost to parm w

	int	GetDelay() const { return edge_delay; }   // return edge delay
	void	SetDelay(int d) { edge_delay = d; };	  // set delay to parm d

	int	GetTotBand(void) const { return total_band; } // return total bandwidth
	void	SetTotBand(int t) { total_band = t; }	      // set total bandwidth to t

	int	GetCurBand(void) const { return current_band; } // return available bandwidth
	void	SetCurBand(int band) { current_band = band; }   // set available bandwidth to parm band
	int	IncCurBand(int inc_band) {
			current_band += inc_band; // increase available bandwidth by parm inc_band
			return current_band;	  // return updated available bandwidth
		}
	int	DecCurBand(int dec_band) {
			current_band -= dec_band; // decrease available bandwidth by parm inc_band
			return current_band; 	  // return updated available bandwidth
		}
	BOOL	IsEnoughBand(int band) {
			if (current_band == UN_LIMIT) return TRUE;
			if (current_band < band) return FALSE;
			return TRUE;
		}
		// return TRUE if there's enough available bandwidth for required band.
		// otherwise, return FALSE.

	UsedBand* 	GetUsedBandHead(void) const { return used_band_head; }
		// get the head of the list of used bandwidth items

	void 	SetUsedBandHead(UsedBand* u) { used_band_head = u; }
		// set the head of list of used bandwidth items to parm u

	void	AddUsedBand(int sess_id, int band);
		// allocate a new item (for session sess_id that requires bandwidth band).
		// add the new item to the front of the list of used bandwidth items.
		// implementation in edge.cc

	int	UpdateUsedBand(ACTION_TYPE type, const UsedBand* u);
	int	UpdateUsedBand(ACTION_TYPE type, const int sess_id, const int band);
		// update the used bandwidth on this edge of session sess_id.
		// the requested or freed bandwidth is band.
		// returns the updated used bandwidth of sess_id.
		// implementation in edge.cc

	STATUS	GetEdgeStatus(void) const { return edge_status; } // return edge status
	void	SetEdgeStatus(STATUS s) { edge_status = s; }	  // set edge status to parm s

	void	UpdateEdge(const Edge* edge);
		// total_band = edge->total_band; current_band = edge->current_band;
		// implementation in edge.cc

	void    DeleteUsedBand();  // delete the whole list of used bandwidth items
				   // implementation in edge.cc
	void	Display(void); 	   // display attributes and list of used bandwidth items
				   // implementation in edge.cc
	// Methods added by UT
	BOOL 	IsEnoughBand_SharedPath(const int req_band, const int sessionID);
	int  	GetSessionUsedBandwidth(const int sessionID);
	int  	UpdateUsedBand_SharedPath(const int sessionID, const int req_band);
	float 	Congestion(int usedBand);
	float 	CongestionInc(int usedBand, int reqBand);
	int	GetEdgeUsedBand() {return (total_band - current_band);}	
};



///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : EdgeItem
//
//	Description : 
//		Wrapper class of Edge class
//		To simplify the list operation of Edge
//
///////////////////////////////////////////////////////////////////////////////
class EdgeItem {
private:
	Edge*		edge;
	EdgeItem*	next;  // to form the list of edges

public:
	EdgeItem() {
		next = NIL;
		edge = NIL;
	}
	EdgeItem(Edge *e) {
		next = NIL;
		edge = e;
	};
	~EdgeItem() { delete edge; }
	Edge* 	GetEdge() const { return edge; };  // return attribute edge
	void	SetEdge(Edge* e) { edge = e; };    // set attribute edge to parm e
	void 	SetNext(EdgeItem* ei) { next = ei; }; // set attribute next to parm ei
	EdgeItem* GetNext() { return next;};	      // return attribute next
};

#endif


///////////////////////////////////////////////////////////////////////////////
//
// Updates:
// Feb 24: added attribute edgeNormFactor to class Edge (UT)
//
///////////////////////////////////////////////////////////////////////////////