///////////////////////////////////////////////////////////////////////////////
//
//	File name : file_io.cc			Last updated: May 7, 98
//
//	Description :
//		This file implements file I/O related methods of class FileIO.
//
//
//	Authors : Jin-Hwee Park / U.T.
// 	Date	: Jan-27-1998
//
///////////////////////////////////////////////////////////////////////////////

#include "file_io.h"


//
// Read input files and generate necessary data structures
//
FileIO::FileIO(DataManager* data, char* dat_file, char* req_file) {
	SetDatFile(dat_file);  // copy file name from parm to attr net_dat_file
	SetReqFile(req_file);  // copy file name from parm to attr grp_req_file	
	data_manager = data;   // set pointer to data_manager object to parm "data"
	FileRead(dat_file, req_file);  	// read input files; genarate network graph,
					// list of sessions, list of request groups
}


//
// Delete dynamic storage
//
FileIO::~FileIO() {
	if (net_dat_file != NIL) delete net_dat_file;
	if (grp_req_file != NIL) delete grp_req_file;
}


//
// Copy file name stored in parm "dat_file" to attribute "net_dat_file"
//
void	FileIO::SetDatFile(char* dat_file) {
	int len = strlen(dat_file);

	net_dat_file = new char[len];
	strcpy(net_dat_file, dat_file);
}


//
// Copy file name stored in parm "req_file" to attribute "grp_req_file"
//
void	FileIO::SetReqFile(char* req_file) {
	int len = strlen(req_file);

	grp_req_file = new char[len];
	strcpy(grp_req_file, req_file);
}


//
// Read the two input files.  Generate the network graph,
// the list of sessions, and the list of request groups.
// Return TRUE if all reads are successful.
//
BOOL	FileIO::FileRead(char* dat_file, char* req_file) {

#ifdef FUNCTION_TRACE
  fprintf(stderr, ".. FileRead() \n");
#endif

	BOOL	ret;
	ret = ReadSystemParms();	// read simulator parameters
	ret = ReadGraphInfo(dat_file);	// read network graph file
	ret = ReadSess_MC(req_file);	// read request file
	return ret;
}


// 
// Read the system parameters from an input file.
// Return TRUE if all reads are successful.
//
BOOL FileIO::ReadSystemParms() 
{
  FILE*	fp;
  float delayFactor, sharedFactor, addedFactor, delayEps;

  if ((fp = fopen(SYS_PARMS_FILE, "r")) == 0)   // file open error
  {
    fprintf(stderr, "FILE[%s], LINE[%d] : %s open error !\n", 
			__FILE__, __LINE__, SYS_PARMS_FILE);
    exit(1);
  }  

  fscanf(fp, "%f", &delayFactor);
  fscanf(fp, "%f", &sharedFactor);
  fscanf(fp, "%f", &addedFactor);
  fscanf(fp, "%f", &delayEps);
  Parameters* parms = new Parameters(delayFactor, sharedFactor, addedFactor, delayEps);
  data_manager->SetParameters(parms);

#ifdef FUNCTION_TRACE
  fprintf(stderr, ".... ReadSystemParms() (de=%.4f, sh=%.4f, ad=%.4f, ep=%.4f) \n",
		delayFactor, sharedFactor, addedFactor, delayEps);
#endif

  fclose(fp);
  return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
//
//	Method name : ReadGraphInfo
//
// 	Description :
//		Read graph information from file and generate network graph
//		(graph object) in the DataManager object.
//
//	Parameter :
//		char* dat_file : network topology file
//
//	Return : TRUE if successful.
//
//	Note : 
//
//	Sample data file => 
//
// 	Node information in the network
//
// 	n: source_node +(: edge_data)
//
// 	edge_data : edge from source_node (start node) to target_node (end node)
// 	target_node, weight, delay, bandwidth
//
// 	If the value of bandwidth is "-1", it is unlimit
//
//	N: 0: 1, 1, 2, 100 : 2, 3, 4, 100 : 3, 5, 1, 50 : 7, 1, 2, 100
//	N: 1: 0, 2, 4, 100 : 2, 4, 5, 50 : 3, 2, 0, 100 : 4, 5, 1, 50 : 
//	7, 5, 7, 100 : 8, 3, 2, 100
//	N: 2: 0, 4, 5, 100 : 1, 4, 2, 50 : 3, 6, 3, 50 : 4, 3,1, 50 : 5, 2, 4, 100
//	N: 3: 0, 4, 5, 100 : 1, 3, 4, 100 :2, 5, 1, 50 : 4, 5, 2, 100
//	N: 4: 2, 3, 1, 50 : 1, 3, 4, 50 : 3, 2, 4, 100
//	N: 5: 2, 2, 4, 100 : 6, 5, 2, 100
//	N: 6: 5, 2, 3, 100
//	N: 7: 0, 5, 2, 100 : 1, 3, 4, 100 : 8, 2, 3, 100
//	N: 8: 1, 3, 4, 100 : 7, 7, 6, 100
//
//	Notes:
//	- a line must start with
//	  + 'N' for a node
//	  + ':' for an adjacent edge
//	- info of an edge (end node, weight, delay, bandwidth) must be on the same line
//
//////////////////////////////////////////////////////////////////////////////
BOOL	FileIO::ReadGraphInfo(char* dat_file)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, ".... ReadGraphInfo() \n");
#endif
	
	DataManager*	d_mgr = GetDataMgr();	// d_mgr = data_manager;
	Graph*	graph = new Graph;		// alloc a Graph object

	FILE*	fp;
	char	sbuf[256];
	int	node_counter=0;

	int		cur_idx;
	READ_ITEM	data;
	Node		*nd1;
	NodeItem	*nd_item;

	if ((fp = fopen(dat_file, "r")) == 0) {  // file open error
		fprintf(stderr, "FILE[%s], LINE[%d] : %s open error !\n", 
			__FILE__, __LINE__, dat_file);
		exit(1);
	}

	while(fgets(sbuf, sizeof(sbuf), fp)) 
	{
	   // Line starts with a new node		
	   if (GetInfoType(sbuf[0]) == NODE_INFO)	// line starts with 'N'
	   {		
		node_counter++;		// one more node
		cur_idx = 0;		// start a new line

		// Read start node and create one Node object
		ReadItem(sbuf, 1, cur_idx, data);	// read node ID (one integer)
		nd_item = new NodeItem;		// alloc new NodeItem object
		nd1 = new Node(data.d1);	// alloc new Node object
		nd_item->SetNode(nd1);		// nd_item->node = nd1;
		graph->InsertNode(nd_item);  	// append nd_item to list of 
						// NodeItems of the graph
		// Read adjacent edges and create an adjacency list
		while(sbuf[cur_idx] != NL) {
			int ret = ReadItem(sbuf, DATA_IN_EDGE, cur_idx, data);
			if (ret > 0 && ret != DATA_IN_EDGE) 
			{
			  printf("FILE[%s], LINE[%d] : format error in network file!\n",
					__FILE__, __LINE__);
			  printf("%s", sbuf);
			  exit(1);
			}  // end if (ret)

			// Create an edge in the adjacency list
			if (ret == DATA_IN_EDGE)	// ret may be 0, hence this test
			{
			  Edge *e = new Edge(nd1, data.d1, data.d2, data.d3, data.d4);
			  nd1->AddAdjEdge(FORWARD, e);  // add "e" to adj list of "nd1"
			}
		}  // end while (sbuf)
	   }  // end if (GetInfoType(sbuf[0]) == NODE_INFO)

	   // Line starts with an adjacent edge of the current node
	   else if (GetInfoType(sbuf[0]) == EDGE_INFO)	// line starts with a digit
	   {
		cur_idx = 0;		// start a new line

		// Continue to read adjacent edges and insert to the adjacency list
		while(sbuf[cur_idx] != NL) {
			int ret = ReadItem(sbuf, DATA_IN_EDGE, cur_idx, data);
			if (ret > 0 && ret != DATA_IN_EDGE) 
			{
			  printf("FILE[%s], LINE[%d] : format error in network file!\n",
					__FILE__, __LINE__);
			  printf("%s", sbuf);
			  exit(1);
			}  // end if (ret)

			// Create an edge in the adjacency list
			if (ret == DATA_IN_EDGE)	// ret may be 0, hence this test
			{
			  Edge *e = new Edge(nd1, data.d1, data.d2, data.d3, data.d4);
			  nd1->AddAdjEdge(FORWARD, e);  // add "e" to adj list of "nd1"
			}
		}  // end while (sbuf)
	   }  // end if (GetInfoType(sbuf[0]) == EDGE_INFO)

	   else continue;	// skip current line if it is not node/edge information				
	}  // end while (fgets)

	fclose(fp);	 // close input file
	graph->SetNodeNum(node_counter);  // graph->num_of_nodes = node_counter;

	// Generate backward adjacency list
	graph->GenBackAdjList();	

	d_mgr->SetGraph(graph);  // d_mgr->graph = graph;
	return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
//
//	Method name : ReadSess_MC
//
// 	Description :
//		Method to read requests, request groups and session information,
//		and generate each data structure in the DataManager object.
//
//	Parameter :
//		char* req_file : data file containing request groups and sessions
//
//	Return : TRUE if successful
//
//	Note : 
//
// 	Sample request file :
//
// 	Session Information
// 	One Session information
// 	S : Session id : Source Node
//
//	SESSIONS {
//	S : 0 : 1
//	S : 1 : 2
//	}
//
// 	MC Group Request
// 	MC Group Request ID
// 	Group Request ID => "I : ID"
// 	User Request => 
// 	"U : Request_node_ID : Session_ID : Req_type : band_width : max_delay : max_hops
// 	Req_type : SETUP(0), JOIN(1), LEAVE(2), TEARDOWN(3)
//
//	GRP_REQ {
//	I : 0
//	U : 0 : 0 : 1 : 10 : 5 : 5
//	U : 1 : 0 : 1 : 10 : 5 : 5
//	U : 2 : 0 : 1 : 10 : 5 : 5
//	U : 3 : 0 : 1 : 10 : 5 : 5
//	U : 3 : 1 : 1 : 10 : 5 : 5
//	U : 4 : 0 : 1 : 10 : 5 : 5
//	U : 5 : 0 : 1 : 10 : 5 : 5
//	U : 6 : 0 : 1 : 10 : 5 : 5
//	U : 7 : 0 : 1 : 10 : 5 : 4
//	U : 8 : 0 : 1 : 10 : 5 : 4
//	}
//
//////////////////////////////////////////////////////////////////////////////
BOOL	FileIO::ReadSess_MC(char* req_file)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, ".... ReadSess_MC() \n");
#endif

	FILE* fp;
	char sbuf[256];
	BOOL	b_sess = FALSE, b_req = FALSE;

	if ((fp = fopen(req_file, "r")) == 0) {     // file open error
		printf("%s open error !\n", req_file);
		exit(1);
	}

        while(fgets(sbuf, sizeof(sbuf), fp)) {
		if (IsSessionInfo(sbuf) == TRUE) ReadSessionInfo(fp);
		else if (IsReqInfo(sbuf) == TRUE) ReadGroupInfo(fp);
	}
	fclose(fp);
	return TRUE;
}


//
// Read input file and build the list of sessions.  Return TRUE if successful.
//
BOOL	FileIO::ReadSessionInfo(FILE* fp)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "...... ReadSessionInfo() \n");
#endif

	char	sbuf[256];
	BOOL	end_of_info = FALSE;

    while(fgets(sbuf, sizeof(sbuf), fp)) {
		if (IsEndOfData(sbuf) == TRUE) {
			end_of_info = TRUE;
			break;
		}
		if (GetInfoType(sbuf[0]) != SESSION_INFO) 
			continue;
		ReadOneSession(sbuf);  // add a new session to list of sessions
	}  

	if (end_of_info == FALSE) {
		printf("FILE[%s], LINE[%d] => Session Information file [%s] format error\n",
			__FILE__, __LINE__, GetReqFile()); 	
		exit(1);
	}
	return TRUE;
}


//
// Read ID of a new group, or a single request.
// Return TRUE if successful.
//
BOOL	FileIO::ReadGroupInfo(FILE* fp)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "...... ReadGroupInfo() \n");
#endif

	GrpReq* mc_req = 0;
	BOOL	end_of_info = FALSE;
	char	sbuf[256];

        while(fgets(sbuf, sizeof(sbuf), fp)) {
		if (IsEndOfData(sbuf) == TRUE) {
			end_of_info = TRUE;
			break;
		}
		
		INFO_TYPE token = GetInfoType(sbuf[0]);
		switch(token) {
			case GRPREQ_ID:
				mc_req = ReadGroupReqID(sbuf);  // create a new group
				break;
			case USER_REQ_INFO:
				ReadUserReq(mc_req, sbuf);  // create a single request
				break;
		}
	}
	if (end_of_info == FALSE || mc_req == 0) {
		printf("FILE[%s], LINE[%d] => Group request Information file [%s] format error\n",
			__FILE__, __LINE__, GetReqFile()); 	
		exit(1);
	}
	return TRUE;
}


//
// IsEndOfData
//
BOOL 	FileIO::IsEndOfData(char* sbuf)
{
	if (sbuf[0] == '}') return TRUE;
	else return FALSE;
}


//
// IsSessionInfo
//
BOOL   	FileIO::IsSessionInfo(char* sbuf)
{
	if (strstr(sbuf, "SESSIONS") != 0) 
		return TRUE;
	else return FALSE;
}


//
// IsReqInfo
//
BOOL   	FileIO::IsReqInfo(char* sbuf)
{
	if (strstr(sbuf, "GRP_REQ") != 0) 
		return TRUE;
	else return FALSE;
}


//
// Read one session; add new Session object to the end of the list of sessions.
// Return TRUE if successful.
// (Session information format "S : Session id : Source Node")
//
BOOL	FileIO::ReadOneSession(char* sbuf)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "........ ReadOneSession() \n");
#endif

	DataManager*	d_mgr = GetDataMgr();  // d_mgr = FileIO->data_manager;
	READ_ITEM	data;
	int	cur_idx=0;

	// Read Session ID
	// Read source node ID
	int ret = ReadItem(sbuf, DATA_IN_SESSION, cur_idx, data); // read from buffer
	if (ret != DATA_IN_SESSION) {
		printf("FILE[%s], LINE[%d] : Session data file format error !\n",
			__FILE__, __LINE__);
		exit(1);
	}
	
	Session* sess = new Session(data.d1, data.d2, d_mgr->GetGraph());
	   // alloc new Session object; session_id = d1; src_node = d2;
	   // MC tree is init using d_mgr->graph (MC tree will have the 
	   // same list of NodeItems as network graph (but adjacency lists
	   // of the nodes are not the same))

	d_mgr->AddSession(sess);  // add new session to end of list of sessions
	return TRUE;
}


//	
// Read a new request group.  
// Append the new request group object to the list of group requests.
// Return the pointer to the new request group object.
// (Group request ID information format "I : ID")
//
GrpReq*	FileIO::ReadGroupReqID(char* sbuf)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "........ ReadGroupReqID() \n");
#endif

	DataManager*	d_mgr = GetDataMgr();  // d_mgr = FileIO->data_manager;
	READ_ITEM	data;
	int	cur_idx=0;

	int ret = ReadItem(sbuf, 1, cur_idx, data);  // read from buffer
	if (ret != 1) {
		printf("FILE[%s], LINE[%d] : Group request data file format error !\n",
			__FILE__, __LINE__);
		exit(1);
	}

	GrpReq* mc_req = new GrpReq(data.d1);  // alloc, init a GrpReq object
	d_mgr->AddGrpReq(mc_req);	// append it to list of group requests
	return mc_req;				// return object pointer 
}


//
// Read a new single request.  Append the new request to the list
// of requests pointed to by parm "mc_grp".
// Return TRUE if successful.
// (User Request information format
// "U : Request_node_ID : Session_ID : Req_type : band_width : max_delay : max_hops")
//
BOOL	FileIO::ReadUserReq(GrpReq* mc_grp, char* sbuf)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "........ ReadUserReq() \n");
#endif

	int	cur_idx=0;
	READ_ITEM	data;

	// Read Request_node_ID 
	// Read Session_ID
	// Read user request type
	// Read bandwidth
	// Read max_delay
	// Read max_hops
	int ret = ReadItem(sbuf, DATA_IN_USER_REQ, cur_idx, data);  // read from buffer
	if (ret != DATA_IN_USER_REQ) {
		printf("FILE[%s], LINE[%d] : Group request data file format error !\n",
			__FILE__, __LINE__);
		exit(1);
	}
	Qos qos(data.d4, data.d5, data.d6);  // alloc & init a new QoS object
	UserReq* user_req = new UserReq(data.d1, data.d2, (USER_REQ_TYPE)data.d3, qos);
						// alloc & init a new UserReq object
	mc_grp->AddUserReq(user_req);	// append new user request to list of 	
					// requests pointed to by "mc_grp"
	return TRUE;
}


//
// GetInfoType
//
INFO_TYPE	FileIO::GetInfoType(char ch)
{
	INFO_TYPE iRet = NOT_INFO;
	switch(ch) {
		case 'N':
			iRet = NODE_INFO;
			break;
		case 'S':
			iRet = SESSION_INFO;
			break;
		case 'I':
			iRet = GRPREQ_ID;
			break;
		case 'U':
			iRet = USER_REQ_INFO;
			break;
		case ':': 
			iRet = EDGE_INFO;
			break;
	}
   	return iRet;
}


//
// IsDigit
//
int	FileIO::IsDigit(char ch)
{
	if (ch >= '0' && ch <= '9' || ch == '-') return TRUE;
	else return FALSE;
}

			
//
//	Read "num_of_items" integers from sbuf[cur_idx] and
//	stores the read values (integers) in structure "data".
// 	Stop when finishing "num_of_items" integers or 
//	encountering an end_of_line.
//
//	Return : the number of integers stored in structure "data"
//
int	FileIO::ReadItem(char* sbuf, int num_of_items, int& cur_idx, READ_ITEM& data)
{
	int 	counter=0;
	char 	sdata[24];

	while(sbuf[cur_idx] != NL) {

		if (IsDigit(sbuf[cur_idx]) == TRUE) {
			int idx = 0;
			while(sbuf[cur_idx] != NL &&
			      IsDigit(sbuf[cur_idx]) == TRUE) {
				sdata[idx] = sbuf[cur_idx];	
				idx++; cur_idx++;
			}  // end if (IsDigit)
			sdata[idx]='\0';
			counter++;
			switch (counter) {
				case 1:
					data.d1 = atoi(sdata);
					break;
				case 2:
					data.d2 = atoi(sdata);
					break;
				case 3:
					data.d3 = atoi(sdata);
					break;
				case 4:
					data.d4 = atoi(sdata);
					break;
				case 5:
					data.d5 = atoi(sdata);
					break;
				case 6:
					data.d6 = atoi(sdata);
					break;
				default:
					printf("FILE[%s], LINE[%d] : Number of read items exceeds the maximum number of storage entries in structure READ_ITEM !\n", 
						__FILE__, __LINE__);
					printf("%s", sbuf);
					exit(1);
					break;
			}  // end switch (counter)
			if (counter ==  num_of_items || sbuf[cur_idx] == NL)
				return counter;
		}  // end if (IsDigit)
		cur_idx++;
	}  // end while (sbuf)
	return counter;
}
