///////////////////////////////////////////////////////////////////////////////
//
//	File name : graph.cc
//
//	Description :
//		This files have graph related class information :
//		Graph, NodeItem and Node classes
//
//
//	Authors	: Jin-Hwee Park / U.T.
// 	Date 	: Jan-25-1998
//
///////////////////////////////////////////////////////////////////////////////

#include "graph.h"

Node::~Node(void) {
	EdgeItem* tmp;

	// Delete forward adjacency list
	while(for_adjlist != NIL) {
		tmp = for_adjlist;
		for_adjlist = for_adjlist->GetNext();
					// for_adjlist = for_adjlist->next
		delete tmp;
	}

	// Delete backward adjacency list
	while(back_adjlist != NIL) {
		tmp = back_adjlist;
		back_adjlist = back_adjlist->GetNext();
					// back_adjlist = back_adjlist->next
		delete tmp;
	}
}

//
// Given the current node, return the head of the fw/bw adjacency list.
//
EdgeItem*	Node::GetEdgeItem(ADJ_TYPE type=FORWARD) const {
	if (type==FORWARD) return for_adjlist;
	else return back_adjlist;
}

//
// Given the current node, display the fw/bw adjacency list
// (list of edges adjacent to current node)
//
void	Node::DisplayEdgeList(ADJ_TYPE type)
{
	EdgeItem* tmp;

	printf("* Node [%d] : ", node_id); // print start node
	if (type == FORWARD) tmp = for_adjlist;   // get head of adj list
	else tmp = back_adjlist;
	while(tmp != NIL) {
		printf("N[%d] : %d,%d,%d -> ",
			tmp->GetEdge()->GetTargetNodeID(),
			tmp->GetEdge()->GetWeight(),
			tmp->GetEdge()->GetDelay(),
			tmp->GetEdge()->GetTotBand());  // tmp->edge->attribute
		tmp = tmp->GetNext();  // tmp = tmp->next
	}
	printf("*\n");
}

//
// Display attributes of object Node
//
void	Node::Display()
{
	printf("* Node INFO : id[%d], for_edge_num[%d], back_edge_num[%d]\n",
		node_id, for_edge_num, back_edge_num);
}

//
//  Add a new edge (parm "edge") to the fw/bw adj list.  Return 1.
//  If the edge is already in the list, do not add.  Just update
//  the edge status if necessary.  Return 0.
//
//  Note: This method can also be used to update the status of an edge in the MC
//	  tree from INTREE to LEAF_NODE (without adding the edge to the tree).
//
int Node::AddAdjEdge(ADJ_TYPE type, Edge *edge)
{
	EdgeItem *tmp;

	if (type == FORWARD) tmp = for_adjlist;  // get head of the list
	else tmp = back_adjlist;

	while (tmp != NIL) {
		int nd_id = tmp->GetEdge()->GetTargetNodeID(); // tmp->edge->target_node_id
		if (nd_id == edge->GetTargetNodeID()) {  // edge->target_node_id
		// edge with "nd_id" is already in the list => do not add it
			if (edge->GetEdgeStatus() == INTREE);  // do nothing
			else if (edge->GetEdgeStatus() == LEAF_NODE) {
				tmp->GetEdge()->SetEdgeStatus(LEAF_NODE);
				  // tmp->edge->edge_status = LEAF_NODE
				  // update status of existing edge to LEAF_NODE
			}
			// delete edge;  // why do this? "edge" is a parm.  See Note 1.
			return 0;
		}
		tmp = tmp->GetNext();  // tmp = tmp->next;
	}

	// Edge with "nd_id" is not in the list.
	// Add it to the front of the list.
	//
	tmp = new EdgeItem(edge);  // alloc a new edge entry
				   // init new entry to parm "edge"
	if (type == FORWARD) {
		tmp->SetNext(for_adjlist);  // tmp->next = for_adjlist;
		for_adjlist = tmp;
	} else {
		tmp->SetNext(back_adjlist);  // tmp->next = back_adjlist;
		back_adjlist = tmp;
	}
	IncEdgeNum(type);  // increment number of edges in the list
	return 1;

};

//
// Note 1: Object "edge" was not added to the graph, and thus unused.  So it should
// be deleted.  However it's not a good practice to delete "edge" here.  This job 
// should be done by the calling method/function, which will check the return value
// (0 or 1).  If the return value is 0, the calling method/function will delete "edge".
//

//
// Increment the number of edges in the fw/bw adj list
//
void	Node::IncEdgeNum(ADJ_TYPE type) 
{
	if (type == FORWARD) 
		for_edge_num++;
	else 
		back_edge_num++;
}

//
// Decrement the number of edges in the fw/bw adj list
//
void	Node::DecEdgeNum(ADJ_TYPE type) 
{
	if (type == FORWARD) 
		for_edge_num--;
	else 
		back_edge_num--;
}

//
// Remove the edge with end node ID "node_id" from the fw/bw adj list.
// Always return 0
//
int Node::RemoveAdjEdge(ADJ_TYPE type, int node_id)
{
	EdgeItem *e_item = NIL;
	EdgeItem *pre_e_item = NIL;

	if (type == FORWARD) e_item = for_adjlist;  // get head of list
	else e_item = back_adjlist;

	while(e_item != NIL) {
		if (e_item->GetEdge()->GetTargetNodeID() == node_id) {
		// (e_item->next->target_node_id (end node) == node_id)
			if (pre_e_item == NIL) {  // e_item is head entry of the list

				//delete e_item;
				//DecEdgeNum(type);
				//if (type == FORWARD)
				//	for_adjlist = 0;
				//else
				//	back_adjlist = 0;
				// the above code (Jin wrote) is wrong!

				if (type == FORWARD)
					for_adjlist = e_item->GetNext();
				else
					back_adjlist = e_item->GetNext();
				delete e_item;
				DecEdgeNum(type);  // decrement number of edges
				break;
			} else {
				pre_e_item->SetNext(e_item->GetNext());
				  // pre_e_item->next = e_item->next
				delete e_item;
				DecEdgeNum(type);  // decrement number of edges
				break;
			}
		}  
		pre_e_item = e_item;
		e_item = e_item->GetNext();  // e_item = e_item->next
	}
	return 0;
}



//////////////////////////////////////////////////////////////////////////////
//
//	Method name : ~Graph
//
// 	Description :
//		Destructor of Graph class.
//		This destroys all Node information in the Graph class
//
//	Parameter :
//
//	Return : 
//
//	Note : 
//
//////////////////////////////////////////////////////////////////////////////
Graph::~Graph() 
{
	NodeItem* tmp;

	// Delete the whole list of nodes of the graph
	while(head != NIL) {
		tmp = head;
		head = head->GetNext();  // head = head->next;
		delete tmp;
	}
}

//
// Traverse the list of nodes; return pointer to the NodeItem whose ID is "id".
// Return NIL pointer if no such NodeItem is found.
//
NodeItem*   Graph::FindNodeItem(int id)
{
	NodeItem*	tmp;

	tmp = head;
	while(tmp != NIL && tmp->GetNode()->GetNodeID() != id)
			 // tmp->node->node_id != id
		tmp = tmp->GetNext();  // tmp = tmp->next;
	if (tmp != NIL) return (tmp);
	return NIL;
}

//
// Traverse the list of nodes; return pointer to the Node whose ID is "id".
// Return NIL pointer if no such Node is found.
//
Node*   Graph::FindNode(int id)
{
	NodeItem*	tmp;

	tmp = FindNodeItem(id);  		// find NodeItem with ID = id
	if (tmp != NIL) return tmp->GetNode();  // return (tmp->node);
	return NIL;
}

//
// Given a node with ID=id, return pointer to the fw/bw adj list of this node.
//
EdgeItem*   Graph::FindEdgeHead(ADJ_TYPE type, int id)
{
	EdgeItem* tmp = NIL;

	tmp = FindNode(id)->GetEdgeItem(type);
		// tmp = current_node.node->for_adjlist (or back_adjlist)
	return tmp;
}

//
// Insert a new NodeItem "e" to the list of nodes.
// The list is ordered in the increasing order of node IDs.
//
void Graph::InsertNode(NodeItem *e) 
{

	int	stop = 0;
	NodeItem *dum;

	e->SetNext(head);  // e->next = head;

	if (head == NIL)
		head = tail = e;
	else if (head->GetNode()->GetNodeID() > e->GetNode()->GetNodeID()) {
	     // (head->node->node_id > e->node->node_id)
		head->SetPrev(e);  // head->prev = e;
		head = e;
	} else while(!stop) {
		dum = e->GetNext();  // dum = e->next;
		e->SetNext(e->GetNext()->GetNext());  // e->next = e->next->next
		if (e->GetNext() == NIL) { // (e->next == NIL)
		// e will be inserted to the end of the list of nodes
			tail->SetNext(e);  // tail->next = e;
			e->SetPrev(tail);  // e->prev = tail;
			tail = e;
			stop = 1;
		} else if (e->GetNext()->GetNode()->GetNodeID() > 
			   e->GetNode()->GetNodeID()) {
			//(e->next->node->node_id > e->node->node_id)
			dum->SetNext(e);  	   // dum->next = e;
			e->SetPrev(dum);  	   // e->prev = dum;
			e->GetNext()->SetPrev(e);  // e->next->prev = e;
			stop = 1;
		};
	};  // end while
	num_of_nodes++;  // increment total number of nodes in the graph
}

//////////////////////////////////////////////////////////////////////////////
//
//	Method name : GenBackAdjList
//
// 	Description :
//		Method to generate backward adjacency list from the existing
//		forward adjacency list.
//		The generated backward adjacency list is used to find the 
//		path from other nodes in the multicasting tree to the request node.
//
//	Parameter :
//
//	Return : void
//
//	Note :
//
//////////////////////////////////////////////////////////////////////////////
void Graph::GenBackAdjList(void)
{
	NodeItem* nd_item_h;

	nd_item_h = head;  // head of list of nodes

	while(nd_item_h != NIL) {  // for every node in the graph
		EdgeItem* edge_h;
		int	start_node_id;

		start_node_id = nd_item_h->GetNode()->GetNodeID();
			   // nd_item_h->node->node_id
		edge_h = nd_item_h->GetNode()->GetEdgeItem(FORWARD);
		      // nd_item_h->node->for_adjlist

		while(edge_h != NIL) {  // for every adj edge of start node
			Edge* org_edge = edge_h->GetEdge();  // edge_h->edge
			Edge* new_edge = new Edge(org_edge);
			  // alloc new Edge object; init attributes to org_edge
			Node* new_start_node;
			// Switch the roles of start node and end node
			int new_src_node_id = org_edge->GetTargetNodeID();
					   // org_edge->target_node_id
			new_edge->SetTargetNodeID(start_node_id);
				// new edge->target_node_id = start_node_id;
			new_start_node = FindNode(new_src_node_id);
				// get pointer to Node from node ID
			new_edge->SetSrcNode(new_start_node);
				// new_edge->source_node = new_start_node;

//printf("[[ Data ]]\n");
//org_edge->Display();
//new_edge->Display();
//exit(1);
			// Add new edge to the backward adj list of
			// new_start_node (old end node)
			new_start_node->AddAdjEdge(BACKWARD, new_edge);
			edge_h = edge_h->GetNext();  // edge_h = edge_h->next
		}
		nd_item_h = nd_item_h->GetNext();  // nd_item_h = nd_item_h->next
	}
}

//
// Given the pointer to start Node object ("nd"), return pointer
// to the adjacency edge whose end node ID is "node_id".
// Return NIL pointer if no such adjacency edge is found.
//
Edge* 	Graph::FindEdge(Node* nd, int node_id)
{
	EdgeItem* edge_item;

	edge_item = nd->GetEdgeItem();  // edge_item = nd->for_adjlist;
	while(edge_item != NIL) {       // for every edge in forward adj list
		if (edge_item->GetEdge()->GetTargetNodeID() == node_id)
		// (edge_item->edge->target_node_id == node_id)
			return edge_item->GetEdge();  // edge_item->edge
		edge_item = edge_item->GetNext();  // edge_item = edge_item->next
	}
	return NIL;  // no such adj edge is found
}

//
// Given the start node ID "s_node", return pointer
// to the adjacency edge whose end node ID is "d_node".
// Return NIL pointer if no such adjacency edge is found.
//
Edge* 	Graph::FindEdge(int s_node, int d_node)
{
	Node*	nd = FindNode(s_node);  // get pointer to Node object from node ID
	return (FindEdge(nd, d_node));  // call the previous method
}

//
// For every node in the graph (list of nodes), display its fw/bw adj list.
// For each edge in the fw/bw adj list, display edge attributes.
//
void 	Graph::Display(ADJ_TYPE type)
{
	NodeItem   *temp;

	printf("*** GRAPH INFO\n");
	temp = head;  // head of list of nodes
	while(temp != NIL) {
		temp->GetNode()->DisplayEdgeList(type);
		temp = temp->GetNext();  // tmp = tmp->next
	}
};

