#ifndef Graph_H
#define Graph_H

#include "common.h"
#include "edge.h"

///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : Node
//
//	Description : 
//		Representation of node in the graph
//		Each node has two adjacency lists for forward, backward
//		- Forward adjacency list is used to maintain the static 
//		  graph information
//		- Backward adjacency list is used to find the shortest
//		  paths from other nodes in the graph to a given node
//
///////////////////////////////////////////////////////////////////////////////
class Node {   
private:
	int		node_id;	// Node ID
	int		for_edge_num;	// Number of edges in the for_adjlist
	int		back_edge_num;	// Number of edges in the back_adjlist

	EdgeItem*   	for_adjlist;   	// Forward adjacency list
	EdgeItem*	back_adjlist;	// Backward adjacency list
public:
	// Constructor
	Node():node_id(-1), for_edge_num(0), for_adjlist(0), back_adjlist(0) {}
	Node(int i) {		// init attributes
		node_id = i;
		for_edge_num = 0;
		back_edge_num = 0;
		for_adjlist = NIL;
		back_adjlist = NIL;
	}
	~Node();

	int	GetNodeID() const { return node_id; };  // return node ID
	void	SetNodeID(int i) { node_id = i; };	// set node ID to parm "i"

	int	GetEdgeNum(ADJ_TYPE type) const {
			if (type == FORWARD) return for_edge_num;
			else return back_edge_num;
		}  // return number of edges in forward/backward list

	void	SetEdgeNum(ADJ_TYPE type, int n_edge) {
			if (type == FORWARD) for_edge_num = n_edge;
			else back_edge_num = n_edge;
		}  // set number of edges in fw/bw list to parm n_edge

	void	IncEdgeNum(ADJ_TYPE type);
		// Increment the number of edges in the fw/bw adj list
		// implementation in graph.cc

	void	DecEdgeNum(ADJ_TYPE type);
		// Decrement the number of edges in the fw/bw adj list
		// implementation in graph.cc

	EdgeItem*  GetEdgeItem(ADJ_TYPE type=FORWARD) const;
	// given current node, return the head of the fw/bw adjacency list
	// implementation in graph.cc

	int	AddAdjEdge(ADJ_TYPE type, Edge *e);
		// Add a new "edge" to the fw/bw adj list.  Return 1.
		// If the edge is already in the list, do not add;
		// just update the edge status if necessary.  Return 0.
		// implementation in graph.cc

	int	RemoveAdjEdge(ADJ_TYPE type, int node_id);
		// remove edge with end node ID "node_id" from fw/bw adj list.
		// always return 0.
		// implementation in graph.cc

	void	Display();  	// display attributes of object Node
				// implementation in graph.cc
	void	DisplayEdgeList(ADJ_TYPE type);
		// given the current node, display the fw/bw adjacency list
		// implementation in graph.cc
}; 


///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : NodeItem
//
//	Description : 
//		Wrapper class for Node class to simplify the list operation
//
///////////////////////////////////////////////////////////////////////////////
class NodeItem { 
private:
	Node* 	node;      
	NodeItem* 	next;  // to form a doubly-linked list of Nodes
	NodeItem* 	prev;  // to form a doubly-linked list of Nodes
public:
	// Constructors & Destructor
	NodeItem() { prev = next = NIL; node = NIL; };
	NodeItem(Node* n) { prev = next = NIL; node =n; };
	~NodeItem() { 
		delete node;
	};

	Node*	GetNode() const { return node; };  // return pointer to a Node
	void	SetNode(Node *n) { node = n; };  // set pointer to a Node to "n"

	NodeItem* GetNext() const { return next; };  // return "next"
	void	SetNext(NodeItem *n) { next = n; };  // set "next" to parm "n"
	NodeItem* GetPrev() const { return prev; };  // return "prev"
	void	SetPrev(NodeItem *n) { prev = n; };  // set "prev" to parm "n"
}; //NodeItem


///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : Graph
//
//	Description : 
//		This class represents the static network topology (graph)
//		information and has a set of methods to maintain the graph
//
//  	Note:
//		The nodes of the graph are represented by a linked list
//		pointed to by a head pointer and a tail pointer.  The list
//		is ordered in the increasing order of node IDs.
//
///////////////////////////////////////////////////////////////////////////////
class Graph { 
private:
	NodeItem *head, *tail;  // head and tail pointer of list of nodes
	int	num_of_nodes;	// total number of nodes in the graph
public:
	Graph() {
		head = tail = NIL;
		num_of_nodes = 0;
	};
	~Graph();

	NodeItem* GetHead() { return(head); };  // return "head" pointer
	NodeItem* GetTail() { return(tail); };  // return "tail" pointer

	int	GetNodeNum(void) { return num_of_nodes; }; // return number of nodes
	void	SetNodeNum(int n) { num_of_nodes = n; };  // set number of nodes to "n"
	void	IncNode(void) { num_of_nodes++; };  // increment number of nodes
	void	DecNode(void) { num_of_nodes--; };  // decrement number of nodes
	
	// Generate backward adjacency list from existing forward adjacency list
	void	GenBackAdjList();

	// Insert a new NodeItem "e" to the list of NodeItems.
	// The list is ordered in the increasing order of node IDs.
	void 	InsertNode(NodeItem *e);   

	// Find the NodeItem in the graph
	// Traverse list of nodes; return pointer to the NodeItem whose ID is "id".
	// Return NIL pointer if no such NodeItem is found.
	NodeItem*	FindNodeItem(int id);

	// Find the real Node object in the graph
	// Traverse list of nodes; return pointer to the Node whose ID is "id".
	// Return NIL pointer if no such Node is found.
	Node*	FindNode(int id);

	// Find the adjacency list pointer of a given Node
	// Given a node with ID=id, return pointer to its fw/bw adj list.
	EdgeItem*	FindEdgeHead(ADJ_TYPE type, int id);

	// Given the pointer to start Node object ("nd"), return pointer
	// to the adjacency edge whose end node ID is "node_id".
	// Return NIL pointer if no such adjacency edge is found.
	Edge*   FindEdge(Node* nd, int node_id);

	// Given the start node ID "s_node", return pointer
	// to the adjacency edge whose end node ID is "d_node".
	// Return NIL pointer if no such adjacency edge is found.
	Edge*   FindEdge(int s_node, int d_node);

	// For every node in the graph, display its fw/bw adj list.
	// For each edge in the fw/bw adj list, display edge attributes.
	void 	Display(ADJ_TYPE type = FORWARD);
}; //Graph

#endif
