///////////////////////////////////////////////////////////////////////////////
//
//	File name : manager.cc			Last updated: Feb 24, 98
//
//	Description :
//		This file implements methods of DataManager class
//
//	Authors : Jin-Hwee Park / U.T.
// 	Date 	: Feb-01-1998
//
///////////////////////////////////////////////////////////////////////////////

#include "manager.h"

//////////////////////////////////////////////////////////////////////////////
//
//	Method name : ~DataManager
//
// 	Description :
//		DataManager class destructor that deletes static
//		and dynamic data structures.
//
//////////////////////////////////////////////////////////////////////////////
DataManager::~DataManager() 
{
	// Delete Parameters object
	if (parameters != NIL) delete parameters;

	// Delete net_env
	if (net_env != NIL) delete net_env;

	// Delete network graph
	if (graph != NIL) delete graph;

	// Delete list of request groups
	GrpReq* tmp;
	while(grp_req_head != NIL) {
		tmp = grp_req_head;
		grp_req_head = grp_req_head->GetNext();
		delete tmp;  // this will delete each object in list of requests
	}

	// Delete list of sessions
	Session* tmp1;
	while(session_head != NIL) {
		tmp1 = session_head;
		session_head = session_head->GetNext();
		delete tmp1;
	}
}

//
// Add a new request group (parm "grp_req") to the end of
// the list of request groups.  Always return 1.
//
int	DataManager::AddGrpReq(GrpReq* grp_req)
{
	GrpReq *tmp0, *tmp1;

	tmp1 = grp_req_head;
	if (tmp1 == NIL)  {
		grp_req_head = grp_req;  // first and only entry
		return 1;
	}
	// tmp0 = tmp1;
	// tmp1=tmp1->GetNext();
	while(tmp1 != NIL) {  // traverse to the end of the list
		tmp0 = tmp1;
		tmp1 = tmp1->GetNext();  // tmp1 = tmp1->next;
	}
	tmp0->SetNext(grp_req);  	 // tmp0->next = grp_req;
	return 1;
}

//
// Add a new session (parm "sess") to the end of the list of sessions.
// Always return 1.
//
int	DataManager::AddSession(Session* sess)
{

	Session *tmp;

	tmp = session_head;
	if (tmp == NIL)  {
		session_head = sess;  // first and only entry
	} else {
		while(tmp->GetNext() != NIL) tmp = tmp->GetNext();
			// traverse to the end of the list of sessions
		tmp->SetNext(sess);  // tmp->next = sess;
	}

	return 1;
}
		
//
// Given session_id, return pointer to the corresponding Session object.
// Return NIL pointer if the Session object is not found.
//
Session*	DataManager::FindSession(int session_id)
{
	Session* sess = GetSessionHead();  // head of list of sessions

	while(sess != NIL) {
		if (sess->GetSessID() == session_id)
			return sess;
		sess = sess->GetNext();  // sess = sess->next;
	}
	return NIL;  // object not found
}

//
// Display all groups of requests
//
void	DataManager::AllReq_Display()
{
	GrpReq* tmp;

	tmp = grp_req_head;  // head of list of request groups
	while(tmp != NIL) {
		tmp->Display();
		tmp = tmp->GetNext();  // tmp = tmp->next;
	}
}

//
// Display all sessions
//
void	DataManager::AllSess_Display()
{
	Session* tmp;

	tmp = session_head;  // head of list of sessions
	while(tmp != NIL) {
		tmp->Display();
		tmp = tmp->GetNext();  // tmp = tmp->next;
	}
}

//
// Display multicast trees of all sessions
//
void	DataManager::AllMCTree_Display()
{
	Session* tmp;

	tmp = session_head;  // head of list of sessions
	while(tmp != NIL) {
		tmp->Display_tree();
		tmp = tmp->GetNext();  // tmp = tmp->next;
	}
}

//
// For every session, display the paths from mcSource
// to the other nodes in the MC tree.
//
void	DataManager::AllPaths_Display()
{
	Session* tmp;

	tmp = session_head;  // head of list of sessions
	while(tmp != NIL) {  // for every session
		// tmp->Display_path();
		tmp->DisplayTreePaths();
		tmp = tmp->GetNext();  // tmp = tmp->next;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
//	Method name : DeleteDynamicData
//
// 	Description :
//		Method to delete every dynamic data previously used and set to
//		initial status
//
//	Parameter :
//
//	Return : void
//
//	Note :
//		This method should be called for each Algorith::RunAlg() method
//		Ex:
//			alg.RunAlg(PROPOSED_ALG);
//			data_mgr->DeleteDynamicData();
//			alg.RunAlg(OTHER_ALG);
//
// 	Note: A better name for this method is ReInitDynamicData
//
//////////////////////////////////////////////////////////////////////////////
void    DataManager::DeleteDynamicData()
{
	// Re-init current and used bandwidths of edges in the graph
	NodeItem* node_head = GetGraph()->GetHead();
	while(node_head != NIL) {
		EdgeItem* edge_head = node_head->GetNode()->GetEdgeItem(FORWARD);
		while(edge_head != NIL) {
			edge_head->GetEdge()->DeleteUsedBand();
			edge_head = edge_head->GetNext();
		}
		node_head = node_head->GetNext();
	}

	// Re-init dynamic data in Session objects
	Session* sess = session_head;  // head of list of sessions
	while(sess != NIL) {
		sess->DeleteDynamicData(GetGraph());  // re-init object "sess"
		sess = sess->GetNext();
	}
}

//
// Display network graph, list of request groups, and list of sessions
//
void	DataManager::Display() {
	printf("\n[[ Parameters ]]\n");
	parameters->Display();	
	printf("\n[[ Forward network information ]]\n");
	graph->Display(FORWARD);
	printf("\n[[ Backward network information ]]\n");
	graph->Display(BACKWARD);
	printf("\n[[ All Sessions ]]\n");
	AllSess_Display();
	printf("\n[[ All Request Groups and Their Requests]]\n");
	AllReq_Display();
}


///////////////////////////////////////////////////////////////////////////////
//
// Updates:
// Feb 24: added class Parameters (UT)
//
///////////////////////////////////////////////////////////////////////////////